@@ -1,3 +1,19 @@
+/*
+ *  This program is free software; you can redistribute it and/or
+ *  modify it under the terms of the GNU General Public
+ *  License as published by the Free Software Foundation; either
+ *  version 2 of the License, or (at your option) any later version.
+ *
+ *  This software is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of  
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *  General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ */
+
+
 #include <gtk/gtk.h>
 
 void
@@ -64,6 +80,10 @@ void
 on_viewContinuous_activate             (GtkMenuItem     *menuitem,
                                         gpointer         user_data);
 
+void
+on_viewHorizontal_activate             (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
 void
 on_viewOnePage_activate                (GtkMenuItem     *menuitem,
                                         gpointer         user_data);
@@ -403,6 +423,16 @@ on_canvas_enter_notify_event           (GtkWidget       *widget,
                                         GdkEventCrossing *event,
                                         gpointer         user_data);
 
+gboolean
+on_canvas_leave_notify_event           (GtkWidget       *widget,
+                                        GdkEventCrossing *event,
+                                        gpointer         user_data);
+
+gboolean
+on_canvas_proximity_event              (GtkWidget       *widget,
+                                        GdkEventProximity *event,
+                                        gpointer         user_data);
+
 gboolean
 on_canvas_expose_event                 (GtkWidget       *widget,
                                         GdkEventExpose  *event,
@@ -435,6 +465,10 @@ void
 on_vscroll_changed                     (GtkAdjustment   *adjustment,
                                         gpointer        user_data);
 
+void
+on_hscroll_changed                     (GtkAdjustment   *adjustment,
+                                        gpointer        user_data);
+
 void
 on_spinPageNo_value_changed            (GtkSpinButton   *spinbutton,
                                         gpointer         user_data);
@@ -471,15 +505,10 @@ void
 on_optionsButtonMappings_activate      (GtkMenuItem     *menuitem,
                                         gpointer         user_data);
 
-void
-on_optionsAntialiasBG_activate         (GtkMenuItem     *menuitem,
-                                        gpointer         user_data);
-
 void
 on_optionsProgressiveBG_activate       (GtkMenuItem     *menuitem,
                                         gpointer         user_data);
 
-
 void
 on_mru_activate                        (GtkMenuItem     *menuitem,
                                         gpointer         user_data);
@@ -598,7 +627,7 @@ on_optionsPrintRuling_activate         (GtkMenuItem     *menuitem,
                                         gpointer         user_data);
 
 void
-on_optionsDiscardCore_activate         (GtkMenuItem     *menuitem,
+on_optionsAutoloadPdfXoj_activate      (GtkMenuItem     *menuitem,
                                         gpointer         user_data);
 
 void
@@ -624,3 +653,52 @@ on_toolsReco_activate                  (GtkMenuItem     *menuitem,
 void
 on_optionsPressureSensitive_activate   (GtkMenuItem     *menuitem,
                                         gpointer         user_data);
+
+void
+on_buttonColorChooser_set              (GtkColorButton  *colorbutton,
+                                        gpointer         user_data);
+
+void
+on_optionsButtonsSwitchMappings_activate(GtkMenuItem    *menuitem,
+                                        gpointer         user_data);
+
+void
+on_toolsImage_activate                 (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
+void
+on_button2Image_activate               (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
+void
+on_button3Image_activate               (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
+void
+on_optionsPenCursor_activate           (GtkCheckMenuItem *checkmenuitem,
+                                        gpointer         user_data);
+
+void
+on_optionsTouchAsHandTool_activate     (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
+void
+on_optionsPenDisablesTouch_activate    (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
+void
+on_optionsDesignateTouchscreen_activate
+                                        (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
+void
+on_journalNewPageKeepsBG_activate      (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
+void
+on_optionsAutosaveXoj_activate         (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
+void
+on_optionsLegacyPDFExport_activate     (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
