@@ -1,3 +1,18 @@
+/*
+ *  This program is free software; you can redistribute it and/or
+ *  modify it under the terms of the GNU General Public
+ *  License as published by the Free Software Foundation; either
+ *  version 2 of the License, or (at your option) any later version.
+ *
+ *  This software is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of  
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *  General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ */
+
 #include <gtk/gtk.h>
 #include <libgnomecanvas/libgnomecanvas.h>
 #include <poppler/glib/poppler.h>
@@ -7,10 +22,10 @@
    and want to list the input events received by xournal. Caution, lots
    of output (redirect to a file). */
 
-#define ENABLE_XINPUT_BUGFIX
-/* comment out this line if you are experiencing calibration problems with
-   XInput and want to try things differently. This will probably break
-   on-the-fly display rotation after application startup, though. */
+// #define ENABLE_XINPUT_BUGFIX
+/* uncomment this line if you are experiencing calibration problems with
+   XInput and want to try things differently. Especially useful on older
+   distributions (up to around 2010). */
 
 #define FILE_DIALOG_SIZE_BUGFIX
 /* ugly, but should help users with versions of GTK+ that suffer from the
@@ -24,6 +39,14 @@
 #define MRU_SIZE 8 
 #define CONFIG_FILE "config"
 
+// version string for about box
+
+#ifdef WIN32
+#define VERSION_STRING VERSION "-win32"
+#else
+#define VERSION_STRING VERSION
+#endif
+
 // DATA STRUCTURES AND CONSTANTS
 
 #define PIXEL_MOTION_THRESHOLD 0.3
@@ -37,6 +60,9 @@
 
 #define VBOX_MAIN_NITEMS 5 // number of interface items in vboxMain
 
+// default touch device
+#define DEFAULT_DEVICE_FOR_TOUCH "Touchscr"
+
 /* a string (+ aux data) that maintains a refcount */
 
 typedef struct Refstring {
@@ -67,7 +93,7 @@ typedef struct Background {
 
 #define BG_SOLID 0
 #define BG_PIXMAP 1
-#define BG_PDF 2      // not implemented yet
+#define BG_PDF 2
 
 #define RULING_NONE 0
 #define RULING_LINED 1
@@ -120,8 +146,9 @@ extern guint predef_bgcolors_rgba[COLOR_MAX];
 #define TOOL_SELECTRECT   5
 #define TOOL_VERTSPACE    6
 #define TOOL_HAND         7
+#define TOOL_IMAGE        8
 #define NUM_STROKE_TOOLS  3
-#define NUM_TOOLS         8
+#define NUM_TOOLS         9
 #define NUM_BUTTONS       3
 
 #define TOOLOPT_ERASER_STANDARD     0
@@ -150,6 +177,10 @@ typedef struct Item {
   gchar *font_name;
   gdouble font_size;
   GtkWidget *widget; // the widget while text is being edited (ITEM_TEMP_TEXT)
+  // the following fields for ITEM_IMAGE:
+  GdkPixbuf *image;  // the image
+  gchar *image_png;  // PNG of original image, for save and clipboard
+  gsize image_png_len;
 } Item;
 
 // item type values for Item.type, UndoItem.type, ui.cur_item_type ...
@@ -178,6 +209,8 @@ typedef struct Item {
 #define ITEM_TEXT_ATTRIB 21
 #define ITEM_RESIZESEL 22
 #define ITEM_RECOGNIZER 23
+#define ITEM_IMAGE 24
+#define ITEM_SELECTREGION 25
 
 typedef struct Layer {
   GList *items; // the items on the layer, from bottom to top
@@ -201,7 +234,7 @@ typedef struct Journal {
 } Journal;
 
 typedef struct Selection {
-  int type;  // ITEM_SELECTRECT, ITEM_MOVESEL_VERT
+  int type;  // ITEM_SELECTRECT, ITEM_MOVESEL_VERT, ITEM_SELECTREGION
   BBox bbox; // the rectangle bbox of the selection
   struct Layer *layer; // the layer on which the selection lives
   double anchor_x, anchor_y, last_x, last_y; // for selection motion
@@ -220,13 +253,17 @@ typedef struct UIData {
   struct Layer *cur_layer;
   gboolean saved; // is file saved ?
   struct Brush *cur_brush;  // the brush in use (one of brushes[...])
-  int toolno[NUM_BUTTONS+1];  // the number of the currently selected tool
+  int toolno[NUM_BUTTONS+2];  // the number of the currently selected tool; two more reserved for eraser tip and touch device
   struct Brush brushes[NUM_BUTTONS+1][NUM_STROKE_TOOLS]; // the current pen, eraser, hiliter
   struct Brush default_brushes[NUM_STROKE_TOOLS]; // the default ones
   int linked_brush[NUM_BUTTONS+1]; // whether brushes are linked across buttons
   int cur_mapping; // the current button number for mappings
   gboolean button_switch_mapping; // button clicks switch button 1 mappings
   gboolean use_erasertip;
+  gboolean touch_as_handtool; // always map touch device to hand tool?
+  gboolean pen_disables_touch; // pen proximity should disable touch device?
+  gboolean in_proximity;
+  char *device_for_touch;
   int which_mouse_button; // the mouse button drawing the current path
   int which_unswitch_button; // if button_switch_mapping, the mouse button that switched the mapping
   struct Page default_page;  // the model for the default page
@@ -244,17 +281,25 @@ typedef struct UIData {
   gboolean pressure_sensitivity; // use pen pressure to control stroke width?
   double width_minimum_multiplier, width_maximum_multiplier; // calibration for pressure sensitivity
   gboolean is_corestroke; // this stroke is painted with core pointer
+  gboolean saved_is_corestroke;
   GdkDevice *stroke_device; // who's painting this stroke
+  gboolean ignore_other_devices;
+  gboolean ignore_btn_reported_up; // config setting: ignore button reported up
+  gboolean current_ignore_btn_reported_up;
   int screen_width, screen_height; // initial screen size, for XInput events
   double hand_refpt[2];
   int hand_scrollto_cx, hand_scrollto_cy;
   gboolean hand_scrollto_pending;
   char *filename;
   gchar *default_path; // default path for new notes
-  gboolean view_continuous, fullscreen, maximize_at_start;
+  gchar *default_image; // path for previous image
+  gboolean fullscreen, maximize_at_start;
+  int view_continuous; // view mode is continuous ?
   gboolean in_update_page_stuff; // semaphore to avoid scrollbar retroaction
   struct Selection *selection;
   GdkCursor *cursor;
+  GdkPixbuf *pen_cursor_pix, *hiliter_cursor_pix;
+  gboolean pen_cursor; // use pencil cursor (default is a dot in current color)
   gboolean progressive_bg; // update PDF bg's one at a time
   char *mrufile, *configfile; // file names for MRU & config
   char *mru[MRU_SIZE]; // MRU data
@@ -262,12 +307,18 @@ typedef struct UIData {
   gboolean bg_apply_all_pages;
   int window_default_width, window_default_height, scrollbar_step_increment;
   gboolean print_ruling; // print the paper ruling ?
+  gboolean exportpdf_prefer_legacy; // prefer legacy code for export-to-pdf?
+  gboolean new_page_bg_from_pdf; // do new pages get a duplicated PDF/image background?
   int default_unit; // the default unit for paper sizes
   int startuptool; // the default tool at startup
   int zoom_step_increment; // the increment in the zoom dialog box
   double zoom_step_factor; // the multiplicative factor in zoom in/out
   double startup_zoom;
   gboolean autoload_pdf_xoj;
+  gboolean autosave_enabled, autosave_loop_running, autosave_need_catchup;
+  GList *autosave_filename_list;
+  int autosave_delay;
+  gboolean need_autosave;
 #if GLIB_CHECK_VERSION(2,6,0)
   GKeyFile *config_data;
 #endif
@@ -282,9 +333,12 @@ typedef struct UIData {
   gboolean shorten_menus; // shorten menus ?
   gchar *shorten_menu_items; // which items to hide
   gboolean is_sel_cursor; // displaying a selection-related cursor
+  gint pre_fullscreen_width, pre_fullscreen_height; // for win32 fullscreen
 #if GTK_CHECK_VERSION(2,10,0)
   GtkPrintSettings *print_settings;
 #endif
+  gboolean poppler_force_cairo; // force poppler to use cairo
+  gboolean warned_generate_fontconfig; // for win32 fontconfig cache
 } UIData;
 
 #define BRUSH_LINKED 0
@@ -300,8 +354,8 @@ typedef struct UndoErasureData {
 
 typedef struct UndoItem {
   int type;
-  struct Item *item; // for ITEM_STROKE, ITEM_TEXT, ITEM_TEXT_EDIT, ITEM_TEXT_ATTRIB
-  struct Layer *layer; // for ITEM_STROKE, ITEM_ERASURE, ITEM_PASTE, ITEM_NEW_LAYER, ITEM_DELETE_LAYER, ITEM_MOVESEL, ITEM_TEXT, ITEM_TEXT_EDIT, ITEM_RECOGNIZER
+  struct Item *item; // for ITEM_STROKE, ITEM_TEXT, ITEM_TEXT_EDIT, ITEM_TEXT_ATTRIB, ITEM_IMAGE
+  struct Layer *layer; // for ITEM_STROKE, ITEM_ERASURE, ITEM_PASTE, ITEM_NEW_LAYER, ITEM_DELETE_LAYER, ITEM_MOVESEL, ITEM_TEXT, ITEM_TEXT_EDIT, ITEM_RECOGNIZER, ITEM_IMAGE
   struct Layer *layer2; // for ITEM_DELETE_LAYER with val=-1, ITEM_MOVESEL
   struct Page *page;  // for ITEM_NEW_BG_ONE/RESIZE, ITEM_NEW_PAGE, ITEM_NEW_LAYER, ITEM_DELETE_LAYER, ITEM_DELETE_PAGE
   GList *erasurelist; // for ITEM_ERASURE, ITEM_RECOGNIZER
@@ -358,6 +412,9 @@ typedef struct BgPdf {
 // the margin between consecutive pages in continuous view
 #define VIEW_CONTINUOUS_SKIP 20.0
 
+#define VIEW_MODE_ONE_PAGE 0 
+#define VIEW_MODE_CONTINUOUS 1
+#define VIEW_MODE_HORIZONTAL 2
 
 // GLOBAL VARIABLES
 
